import os

from contextlib import contextmanager
from itertools import zip_longest

import click
from jinja2 import FileSystemLoader, TemplateNotFound


def iterable_converged(left, right):
    """
    Returns (True, None) if the two iterables generate identical,
    otherwise returns (False, index) where index indicates the first position where
    the iterables differ.
    """
    for i, (l_item, r_item) in enumerate(zip_longest(left, right)):
        if l_item != r_item:
            return False, i
    return True, None


class RestrictedFileSystemLoader(FileSystemLoader):
    def get_source(self, environment, template):
        self._ensure_not_unsafe_github(template)
        self._ensure_not_git(template)

        return super().get_source(environment, template)

    def list_templates(self):
        def only_safe(template):
            try:
                self._ensure_not_git(template)
                self._ensure_not_unsafe_github(template)
                return True
            except TemplateNotFound:
                return False

        return filter(only_safe, super().list_templates())

    @staticmethod
    def _ensure_not_unsafe_github(template):
        if template.startswith(".github/") and not (
            template.endswith(".ght") or template.endswith(".j2")
        ):
            raise TemplateNotFound(
                f"Templates under the .github/ folder must end in .ght or j2: {template}"
            )

    @staticmethod
    def _ensure_not_git(template):
        if template.startswith(".git/"):
            raise TemplateNotFound(f"The .git folder is not a valid path for templates: {template}")


@contextmanager
def stashed_checkout(repo, branch_name):
    with stashed(repo) as stash:
        with checkout(repo, branch_name) as ref:
            yield stash, ref


@contextmanager
def stashed(repo):
    """A context that stashes all uncommitted/untracked items"""
    prev_stashed_items = len(repo.git.stash("list").splitlines())
    repo.git.stash("push", "--all", "-m", "[ght]: Before editing configuration file.")
    curr_num_stashed_items = len(repo.git.stash("list").splitlines())

    stash_created = curr_num_stashed_items - prev_stashed_items > 0

    yield stash_created

    if stash_created:
        repo.git.stash("pop")


@contextmanager
def checkout(repo, branch_name):
    """Branch checkout context"""
    prev_head = repo.head.ref
    yield repo.heads[branch_name].checkout()
    prev_head.checkout()


def resolve_repository_path(repo_path):
    # Find the configuration file up the directory tree
    while not os.path.isfile(f"{repo_path}/.github/ght.yaml"):
        parent_dir = os.path.dirname(repo_path)
        if parent_dir == repo_path:
            raise click.UsageError(
                "Not a gittr repository (or any of the parent directories): .github/ght.yaml"
            )
        repo_path = parent_dir
    return repo_path



import os
from contextlib import contextmanager

import yaml
from platform import release

from gittr.cli.utils import iterable_converged, RestrictedFileSystemLoader
from jinja2 import Environment, Template
from git import Repo, GitConfigParser


class GHT(object):
    repo: Repo
    env: Environment
    config: dict
    template_url: str
    template_ref: str

    __slots__ = ["repo", "env", "config", "config_path", "template_url", "template_ref"]

    def __init__(self, repo_path, template_url=None, template_ref="master", config_path=None):
        self.repo = Repo(path=repo_path)
        self.template_url = template_url
        self.template_ref = template_ref
        self.config_path = config_path or os.path.join(
            self.repo.working_tree_dir, ".github", "ght.yaml"
        )

        self.env = Environment(
            loader=RestrictedFileSystemLoader(self.repo.working_tree_dir),
            extensions=[
                "jinja2.ext.do",
                "jinja2.ext.loopcontrols",
                "jinja2.ext.with_",
                "jinja2_time.TimeExtension",
            ],
        )
        self.configure_author()

    def load_config(self):
        if not os.path.exists(self.config_path):
            raise ValueError(
                f"{self.repo.working_tree_dir} is an invalid GHT repository, "
                "{self.config_path} does not exist."
            )
        with open(self.config_path, "r") as f:
            self.config = yaml.load(f, Loader=yaml.SafeLoader)
            self.template_url = self.template_url or self.config["ght"]["template"]["url"]

    def configure_author(self):
        """
        git config --local user.email ""
        git config --local user.name ""
        """
        cw: GitConfigParser
        with self.repo.config_writer() as cw:
            cw.set_value("user", "email", "psodre@gmail.com")
            cw.set_value("user", "name", "Patrick Sodré")
        # call release() to be sure changes are written and locks are released
        release()

    def prepare_tree_for_rendering(self):
        """
        git rm -rf .
        git checkout ght/template -- .
        git branch -D ght/template
        git checkout HEAD -- .github/ght.yaml
        """
        self.remove_all()

        with self.fetch_template():
            self.repo.git.checkout("ght/template", "--", ".")

        self.repo.git.checkout("HEAD", "--", ".github/ght.yaml")

    @contextmanager
    def fetch_template(self):
        self.repo.git.fetch(self.template_url, "--no-tags", f"{self.template_ref}:ght/template")
        yield
        self.repo.git.branch("-D", "ght/template")

    def remove_all(self):
        """
        Does the equivalent of `git rm -rf .`
        """
        all_blobs = [
            o.path
            for o in self.repo.tree().traverse(
                predicate=lambda i, _: i.type == "blob", branch_first=False
            )
        ]
        for path in all_blobs:
            fs_path = os.path.join(self.repo.working_tree_dir, path)
            os.remove(fs_path)
        self.repo.index.remove(all_blobs)
        all_trees = [
            o.path
            for o in self.repo.tree().traverse(
                predicate=lambda i, _: i.type == "tree", branch_first=False
            )
        ]
        all_trees.reverse()
        for path in all_trees:
            os.rmdir(os.path.join(self.repo.working_tree_dir, path))
        self.repo.index.update()

    def render_ght_conf(self):
        """
        Render the .github/ght.yaml file
        """
        ght_conf_path = os.path.join(self.repo.working_tree_dir, ".github", "ght.yaml")
        with open(ght_conf_path) as f:
            curr_ght_yaml = f.read().splitlines()
        next_ght_yaml = curr_ght_yaml

        converged, index = False, -1
        while not converged:
            curr_ght_yaml = next_ght_yaml[: index + 1] + curr_ght_yaml[index + 1 :]  # noqa: E203
            config = yaml.safe_load("\n".join(curr_ght_yaml))
            next_ght_yaml = [self.env.from_string(line).render(config) for line in curr_ght_yaml]
            converged, index = iterable_converged(curr_ght_yaml, next_ght_yaml)

        with open(ght_conf_path, "w") as f:
            f.write("\n".join(curr_ght_yaml))
        self.repo.index.add(".github/ght.yaml")

    def render_tree(self):
        self.prepare_tree_for_rendering()
        self.render_ght_conf()
        self.load_config()
        self.render_tree_content()
        self.repo.index.commit(f"[ght]: rendered {self.template_url} content", skip_hooks=True)
        self.render_tree_structure()
        self.repo.index.commit(f"[ght]: rendered {self.template_url} structure", skip_hooks=True)

    def render_tree_structure(self):
        """
        Renders the Tree structure in git, by applying `render_ght_obj_name` to each object name.
        """
        objs_to_rename = [
            (o.path, os.path.join(os.path.dirname(o.path), new_name))
            for o in self.repo.tree().traverse(branch_first=False)
            if o.name != (new_name := self.render_ght_obj_name(o.name))
        ]
        objs_to_rename.reverse()

        for old_new in objs_to_rename:
            self.repo.index.move(old_new)
            self.repo.index.update()

    def render_ght_obj_name(self, name):
        if name.endswith(".ght"):
            rv = name[:-4]
        else:
            rv = name
        return self.env.from_string(rv).render(self.config)

    def render_tree_content(self):
        """
        Render all tree content
        """
        paths_to_render = [
            o.path
            for _, o in self.repo.index.iter_blobs()
            if not o.path.startswith(".github/") or o.path.endswith(".ght")
        ]

        for path in paths_to_render:
            template: Template = self.env.get_template(path)
            rendered = template.render(self.config)
            with open(os.path.join(self.repo.working_tree_dir, path), "w") as f:
                f.write(rendered)
            self.repo.index.add(path)

    @classmethod
    def init(cls, path, config: dict = None, **kwargs):
        """
        Step 1: Initialize the git repo
        Step 2: Write the configuration file to master branch
        Step 3: Create the ght/master branch from master
        """
        # Step 1: Initialize the git repo
        repo = Repo.init(path)

        ght = cls(repo_path=path, **kwargs)

        # Step 2: Write the configuration file to the master branch
        #   if config is a dict, then yaml.dump it
        if config is None:
            # Get the configuration file from the template URL
            with ght.fetch_template():
                repo.git.checkout("ght/template", ".github/ght.yaml")
        elif isinstance(config, dict):
            github_dir = os.path.join(path, ".github")
            os.makedirs(github_dir, exist_ok=True)
            with open(os.path.join(github_dir, "ght.yaml"), "w") as f:
                yaml.dump(config, f)
            repo.index.add(".github/ght.yaml")
        else:
            raise ValueError("config must be None or a dictionary.")
        repo.index.commit("[ght]: Add ght.yaml configuration.", skip_hooks=True)
        ght.load_config()

        # Step 3: Create the ght/master branch
        repo.create_head("ght/master")

        return ght



"""Console script for gittr."""

import collections
import os

import click
from click_plugins import with_plugins
from entrypoints import get_group_named

from gittr.cli.action import GHT
from gittr.cli.utils import stashed_checkout, resolve_repository_path


class OrderedGroup(click.Group):
    """A click group that maintains order.
    ref: https://bit.ly/click-ordered-group
    """

    def __init__(self, name=None, commands=None, **attrs):
        super(OrderedGroup, self).__init__(name, commands, **attrs)
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands


@with_plugins(get_group_named("gittr").values())
@click.group(cls=OrderedGroup)
def cli():
    """gittr command-line-interface"""
    return 0


@cli.command("init")
@click.argument("repository", type=str)
@click.argument("refspec", type=str, default="master", metavar="[REFSPEC]")
def init(repository, refspec):
    """Initialize a git project from a ght-template repository.

    \b
    REPOSITORY: A git-url pointing to a ght-repository to use as a template.
    REFSPEC: The git refspec to use. [default: master]

    \b
    This command will:
      - Git-initialize the current working directory
      - Download the ght template configuration file
      - Create the ght/master tracking branch

    \b
    EXAMPLES:
        $ mkdir example
        $ cd example
        $ ght init https://github.com/sodre/ght-pypackage master
    """

    if len(os.listdir(".")) > 0:
        raise click.ClickException("The current directory is not empty, refusing to initialize it.")

    # Setup the GHT Repository
    _ = GHT.init(path=".", template_url=repository, template_ref=refspec)

    return 0


@cli.command("configure")
@click.argument("repo-path", default=".", type=click.Path(file_okay=False, exists=True))
def configure(repo_path):
    """Edit an existing template configuration file

    A git commit is created if the file is modified.
    """

    # Open the repo
    repo_path = resolve_repository_path(repo_path)
    ght = GHT(repo_path, None)

    with stashed_checkout(ght.repo, "ght/master"):
        click.edit(filename=f"{repo_path}/.github/ght.yaml")
        ght.repo.index.add(".github/ght.yaml")
        ght.repo.index.commit("[ght]: Update configuration file.", skip_hooks=True)


@cli.command()
@click.argument("refspec", default="master", metavar="[REFSPEC]")
@click.argument("dest-branch", default="ght/master", metavar="[GHT_BRANCH]")
def render(refspec, dest_branch):
    """(Re)render the project

    \b
    REFSPEC: The template branch/refspec to use for rendering [default=master]
    GHT_BRANCH: The destination branch of the rendered results [default=ght/master]
    """
    if not dest_branch.startswith("ght/"):
        raise click.ClickException(
            "Refusing to render the template."
            f"The destination branch `{dest_branch}` does not begin ght/."
        )

    repo_path = resolve_repository_path(".")
    ght = GHT(repo_path=repo_path, template_ref=refspec)

    with stashed_checkout(ght.repo, dest_branch):
        ght.load_config()
        ght.render_tree()

    return 0


@cli.command("approve")
@click.argument("commit", default="ght/master")
def approve(commit):
    """Merge the rendered template from ght/master to master
    """

    repo_path = resolve_repository_path(".")
    ght = GHT(repo_path=repo_path)

    with stashed_checkout(ght.repo, "master"):
        click.echo(ght.repo.git.merge("--no-squash", "--no-ff", commit))
